(*)___________________________________________________________________

  Project:

      MediaTag
      ________

      (C) 2004 Maxim Proskurnya,
               max@axofiber.org.ru

  Component:  fLaCPArserUnit
  Comments :  Implementation of the OGG-Vorbis tags read-write
              operations in flac media files.
  ____________________________________________________________________

  17.XI.2004                                                     00.01
    Fixed value of the vorbis comment block size.  
  03.XI.2004                                                     00.00
   ___________________________________________________________________
(*)

unit fLaCParserUnit;

interface

uses
  Classes;

type

  TfLaCParserResult  = ( fLaC_OK, fLaC_FILE_NOT_FOUND, fLaC_BAD_FORMAT );

  TfLaCMetaBlockType = ( fLaC_STREAMINFO,
                         fLaC_PADDING,
                         fLaC_APPLICATION,
                         fLaC_SEEKTABLE,
                         fLaC_VORBIS_COMMENT,
                         fLaC_CUESHEET,
                         fLaC_RESERVED );


  CharFile = File of Char;

  TfLaCParser = class
    function ReadTags  ( FileName: string; var SampleRate, Channels, BPS: Integer; var TotalSamplesQnt, MediaFileSize: Int64; Comments: TStrings ): TfLaCParserResult;
    function WriteTags ( FileName: string; Comments: TStrings ): TfLaCParserResult;
  private
    MediaFile : CharFile;
    function ReadMagicBytes: boolean;
    function ReadBlockHeader ( var IsFinal: Boolean; var BlockType : TfLaCMetaBlockType; var Length: Integer ): boolean;
    function ReadStreamInfo ( var SampleRate, Channels, BPS: Integer; var TotalSamplesQnt: Int64 ): Boolean;
    function ReadVorbisComment ( Comments: TStrings ): Boolean;
  end;

implementation

uses
  Sysutils;



  function ConvertToUTF8 ( s: string ): string;
  var
    i : Integer;
  begin
    Result := '';
    i := 1;
    while i <= Length ( s ) do
    begin
      if ord ( s[i] ) < $80 then
        Result := Result + s[i]
      else if s[i] = '' then
        Result := Result + chr ( $D0 ) + chr ( $81 )
      else if s[i] = '' then
        Result := Result + chr ( $D1 ) + chr ( $91 )
      else if ( $C0 <= ord ( s[i] ) ) and ( ord ( s[i] ) <= $EF ) then
        Result := Result + chr ( $D0 ) + chr ( ord ( s[i] ) - $30 )
      else if $F0 <= ord ( s[i] ) then
        Result := Result + chr ( $D1 ) + chr ( ord ( s[i] ) - $70 );
      Inc ( i );
    end;
  end;



  function ConvertFromUTF8 ( s: string ): string;
  var
    i : Integer;
  begin
    Result := '';
    i := 1;
    while i <= Length ( s ) do
    begin
      if ord ( s[i] ) < $80 then
        Result := Result + s[i]
      else if s[i] = chr ( $D0 ) then
      begin
        Inc ( i );
        if s[i] = chr ( $81 ) then
          Result := Result + ''
        else
          Result := Result + chr ( ord ( s[i] ) + $30 )
      end
      else if s[i] = chr ( $D1 ) then
      begin
        Inc ( i );
        if s[i] = chr ( $91 ) then
          Result := Result + ''
        else
          Result := Result + chr ( ord ( s[i] ) + $70 )
      end;
      Inc ( i );
    end;
  end;



  function TfLaCParser.ReadBlockHeader ( var IsFinal: Boolean; var BlockType : TfLaCMetaBlockType; var Length: Integer ): boolean;
  var
    c    : Char;
    Data : Longint;
  begin
    Result := False;
    if not eof ( MediaFile ) then
    begin
      Read ( MediaFile, c );
      Data    := ord ( c );
      IsFinal := ( Data and 128 ) <> 0;
      case Data and 127 of
        0 :  BlockType := fLaC_STREAMINFO;
        1 :  BlockType := fLaC_PADDING;
        2 :  BlockType := fLaC_APPLICATION;
        3 :  BlockType := fLaC_SEEKTABLE;
        4 :  BlockType := fLaC_VORBIS_COMMENT;
        5 :  BlockType := fLaC_CUESHEET;
        else BlockType := fLaC_RESERVED;
      end;
      if not eof ( MediaFile ) then
      begin
        Read ( MediaFile, c );
        Data := ord ( c );
        if not eof ( MediaFile ) then
        begin
          Read ( MediaFile, c ); Data := Data shl 8 + ord ( c );
          if not eof ( MediaFile ) then
          begin
            Read ( MediaFile, c ); Data := Data shl 8 + ord ( c );
            Length := Data;
            Result := True;
          end;
        end;
      end;
    end;
  end;



  function TfLaCParser.ReadMagicBytes: boolean;
  var
    i : Integer;
    s : string;
    c : char;
  begin
    i := 0;
    while ( i < 4 ) and not eof ( MediaFile ) do
    begin
      Read ( MediaFile, c );
      s := s + c;
      Inc ( i );
    end;
    if s = 'fLaC' then
      Result := True
    else
      Result := False;
  end;



  function TfLaCParser.ReadStreamInfo ( var SampleRate, Channels, BPS: Integer; var TotalSamplesQnt: Int64 ): Boolean;
  var
    c : Char;
  begin
    Result := False;
    if not eof ( MediaFile ) then
    begin
      Seek ( MediaFile, FilePos ( MediaFile ) + 10 );
      Read ( MediaFile, c ); SampleRate := ord ( c );
      Read ( MediaFile, c ); SampleRate := SampleRate shl 8 + ord ( c );
      Read ( MediaFile, c ); SampleRate := SampleRate shl 4 + ( ord ( c ) shr 4 );
      Channels := ( ord ( c ) shr 1 ) and 7 + 1;
      BPS := ord ( c ) and 1;
      if not eof ( MediaFile ) then
      begin
        Read ( MediaFile, c );
        BPS := BPS shl 4 + ( ord ( c ) shr 4 ) + 1;
        TotalSamplesQnt := ord ( c ) and 15;
        Read ( MediaFile, c );
        TotalSamplesQnt := TotalSamplesQnt shl 8 + ord ( c );
        Read ( MediaFile, c );
        TotalSamplesQnt := TotalSamplesQnt shl 8 + ord ( c );
        Read ( MediaFile, c );
        TotalSamplesQnt := TotalSamplesQnt shl 8 + ord ( c );
        Read ( MediaFile, c );
        TotalSamplesQnt := TotalSamplesQnt shl 8 + ord ( c );
        Seek ( MediaFile, FilePos ( MediaFile ) + 16 );
        Result := True;
      end;
    end;
  end;



  function TfLaCParser.ReadTags ( FileName: string; var SampleRate, Channels, BPS: Integer; var TotalSamplesQnt, MediaFileSize: Int64; Comments: TStrings ): TfLaCParserResult;
  var
    IsFinal   : Boolean;
    BlockType : TfLaCMetaBlockType;
    Length    : Integer;
  begin
    if FileExists ( FileName ) then
    begin
      AssignFile ( MediaFile, FileName );
      Reset ( MediaFile );
      MediaFileSize := FileSize ( MediaFile );
      Result := fLaC_BAD_FORMAT;
      if ReadMagicBytes then
        if ReadBlockHeader ( IsFinal, BlockType, Length ) and ( BlockType = fLaC_STREAMINFO ) then
        begin
          ReadStreamInfo ( SampleRate, Channels, BPS, TotalSamplesQnt );
          while ReadBlockHeader ( IsFinal, BlockType, Length ) do
          begin
            if BlockType = fLaC_VORBIS_COMMENT then
            begin
              Comments.Clear;
              if ReadVorbisComment ( Comments ) then
                Result := fLaC_OK;
            end
            else
              Seek ( MediaFile, FilePos ( MediaFile ) + Length );
            iF IsFinal then Break;
          end;
        end;
      MediaFileSize := FileSize ( MediaFile ) - FilePos ( MediaFile );
      CloseFile ( MediaFile );
    end
    else
      Result := fLaC_FILE_NOT_FOUND;
  end;



  function TfLaCParser.ReadVorbisComment ( Comments: TStrings ): Boolean;
  var
    c       : Char;
    Data    : Longint;
    Comment : string;
    i, j, n : Longint;
    EqPos   : Integer;
  begin
    Result := False;
    if not eof ( MediaFile ) then
    begin
      // vendor_string length
      Read ( MediaFile, c ); Data := ord ( c );
      Read ( MediaFile, c ); Data := ord ( c ) shl 8  + Data;
      Read ( MediaFile, c ); Data := ord ( c ) shl 16 + Data;
      Read ( MediaFile, c ); Data := ord ( c ) shl 24 + Data;
      // skip vendor_string
      Seek ( MediaFile, FilePos ( MediaFile ) + Data );
      // number of comments
      Read ( MediaFile, c ); Data := ord ( c );
      Read ( MediaFile, c ); Data := ord ( c ) shl 8 + Data;
      Read ( MediaFile, c ); Data := ord ( c ) shl 16 + Data;
      Read ( MediaFile, c ); n    := ord ( c ) shl 24 + Data;
      for i := 1 to n do
      begin
        // comment length
        Read ( MediaFile, c ); Data := ord ( c );
        Read ( MediaFile, c ); Data := ord ( c ) shl 8 + Data;
        Read ( MediaFile, c ); Data := ord ( c ) shl 16 + Data;
        Read ( MediaFile, c );
        Comment := '';
        for j := 1 to ord ( c ) shl 24 + Data do
        begin
          Read ( MediaFile, c );
          Comment := Comment + c;
        end;
        Comment := Trim ( COnvertFromUTF8 ( Comment ) );
        if 1 < Length ( Comment ) then
        begin
          EqPos := Pos ( '=', Comment );
          if 1 < EqPos then
          Comments.Values[ UpperCase ( Copy (Comment, 1, EqPos - 1 ) ) ] := Copy ( Comment, EqPos + 1, Length ( Comment ) - EqPos );
        end;
      end;
      Result := True;
    end;
  end;



  function TfLaCParser.WriteTags ( FileName: string; Comments: TStrings ): TfLaCParserResult;
  var
    IsFinal       : Boolean;
    BlockType     : TfLaCMetaBlockType;
    SampleRate, Channels, BPS: Integer;
    TotalSamplesQnt          : INt64;
    i, BlockLength  : Longint;
    VorbisBlockSize : Longint;
    Gap             : Longint;
    StartOfComment, EndOfComment, StartOfPadding, PaddingLength: Longint;
    c               : Char;
    VendorName      : string;


    procedure  MoveBytes ( var MediaFile: CharFile; Source, Dest, SizeInBytes : LongInt );
    var
      i : Integer;
      c : Char;
    begin
      if ( SizeInBytes < abs ( Dest - Source ) ) or
         ( Dest < Source  )
      then
        for i := 0 to SizeInBytes - 1 do
        begin
          Seek  ( MediaFile, Source + i );
          Read  ( MediaFile, c );
          Seek  ( MediaFile, Dest + i );
          Write ( MediaFile, c );
        end
      else
        for i := SizeInBytes - 1 downto 0 do
        begin
          Seek  ( MediaFile, Source + i );
          Read  ( MediaFile, c );
          Seek  ( MediaFile, Dest + i );
          Write ( MediaFile, c );
        end;
    end;


    procedure WriteBuf ( Buf: string );
    var
      L, i : Integer;
    begin
      L := Length ( Buf );
      for i := 1 to L do
        Write ( MediaFile, Buf[i] );
    end;

    procedure WriteIntR ( Number: Int64; BytesSize: integer );
    var
      i : Integer;
      c : Char;
    begin
      for i := 1 to BytesSize do
      begin
        c := chr ( Number and 255 );
        Write ( MediaFile, c );
        Number := Number shr 8;
      end;
    end;

    procedure WriteInt ( Number: Int64; BytesSize: integer );
    var
      i     : Integer;
      Shift : Integer;
      c     : Char;
    begin
      Shift := ( BytesSize - 1 ) * 8;
      for i := 1 to BytesSize do
      begin
        c := chr ( Number shr Shift and 255 );
        Shift := Shift - 8;
        Write ( MediaFile, c );
      end;
    end;


  begin
    if FileExists ( FileName ) then
    begin
      PaddingLength := 0;
      AssignFile ( MediaFile, FileName );
      Reset ( MediaFile );
      Result := fLaC_BAD_FORMAT;
      EndOfComment   := 0;
      StartOfComment := 0;
      StartOfPadding := 0;
      if ReadMagicBytes then
        if ReadBlockHeader ( IsFinal, BlockType, BlockLength ) and ( BlockType = fLaC_STREAMINFO ) then
        begin
          ReadStreamInfo ( SampleRate, Channels, BPS, TotalSamplesQnt );
          while ReadBlockHeader ( IsFinal, BlockType, BlockLength ) do
          begin
            if BlockType = fLaC_VORBIS_COMMENT then
            begin
              // Store the vorbis-comment block start and final position
              StartOfComment := FilePos ( MediaFile ) - 4;
              EndOfComment   := FilePos ( MediaFile ) + BlockLength;
            end
            else if BlockType = fLaC_PADDING then
            begin
              // Found padding area which will help to picth the metablocks
              StartOfPadding := FilePos ( MediaFile ) - 4;
              PaddingLength  := BlockLength;
            end;
            if IsFinal then
              Break
            else
              Seek ( MediaFile, FilePos ( MediaFile ) + BlockLength );
          end;
        end;
      if 0 < PaddingLength then
      begin
        VendorName := 'reference libFLAC 1.1.0 20030126';
        // Now able to write new comments (when )

        // Calc new block size
        // Blocksize + VendorStrLength + VendorString + CommentsQnt
        VorbisBlockSize := 4 + 4 + Length ( VendorName ) + 4;

        for i := 0 to Comments.Count - 1 do
        begin
          Comments.Strings[i] := ConvertToUTF8 ( Comments.Strings[i] );
//S := ConvertFromUTF8 ( Comments.Strings[i] );
          VorbisBlockSize := VorbisBlockSize + 4 + Length ( Comments.Strings[i] );
        end;

        Gap := VorbisBlockSize - EndOfComment + StartOfComment;

        if Gap <> 0 then
        begin
          // Shrunk vorbis block
          //                     from          to                  size
          MoveBytes ( MediaFile, EndOfComment, EndOfComment + Gap, abs ( Gap ) )
        end;

        // Write comments header
        Seek ( MediaFile, StartOfComment );
        WriteBuf ( chr ( ord ( fLaC_VORBIS_COMMENT ) ) );
        WriteInt ( VorbisBlockSize - 4, 3 );
        WriteIntR( Length ( VendorName ), 4 );
        WriteBuf ( VendorName );
        // number of comments
        WriteIntR ( Comments.Count, 4 );
        for i := 0 to Comments.Count - 1 do
        begin
          WriteIntR ( Length ( Comments.Strings[i] ), 4 );
          WriteBuf ( Comments.Strings[i] );
        end;

        if Gap <> 0 then
        begin
          // Pitch padding
          Seek ( MediaFile, StartOfPadding + Gap );
          WriteBuf ( chr ( 1 shl 7 + ord ( fLaC_PADDING ) ) );
          WriteInt ( StartOfPadding + PaddingLength + Gap, 3 );
          if Gap < 0 then
          begin
            c := #0;
            for i := 1 to -Gap do
              // Fill by null-bytes
              Write ( MediaFile, c );
          end;
        end
      end;
      CloseFile ( MediaFile );
    end
    else
      Result := fLaC_FILE_NOT_FOUND;
  end;

end.
